package com.example.meshtest

import android.bluetooth.BluetoothDevice
import android.bluetooth.BluetoothGatt
import android.bluetooth.BluetoothGattCharacteristic
import android.bluetooth.le.ScanResult
import android.content.Context
import android.os.ParcelUuid
import android.util.Log
import no.nordicsemi.android.ble.BleManager
import no.nordicsemi.android.ble.callback.DataReceivedCallback
import no.nordicsemi.android.ble.callback.DataSentCallback
import no.nordicsemi.android.ble.callback.SuccessCallback
import no.nordicsemi.android.ble.data.Data
import no.nordicsemi.android.mesh.MeshManagerApi
import no.nordicsemi.android.mesh.MeshManagerCallbacks
import no.nordicsemi.android.mesh.MeshNetwork
import no.nordicsemi.android.mesh.MeshProvisioningStatusCallbacks
import no.nordicsemi.android.mesh.MeshStatusCallbacks
import no.nordicsemi.android.mesh.UnprovisionedBeacon
import no.nordicsemi.android.mesh.provisionerstates.ProvisioningState
import no.nordicsemi.android.mesh.provisionerstates.UnprovisionedMeshNode
import no.nordicsemi.android.mesh.transport.ControlMessage
import no.nordicsemi.android.mesh.transport.MeshMessage
import no.nordicsemi.android.mesh.transport.ProvisionedMeshNode
import java.util.UUID

class Mesh(context: Context): BleManager(context), MeshManagerCallbacks,
    MeshProvisioningStatusCallbacks, MeshStatusCallbacks {
        companion object {
            val MTU_SIZE_DEFAULT: Int = 23
            val MTU_SIZE_MAX: Int = 517

            /**
             * Mesh provisioning service UUID
             */
            val MESH_PROVISIONING_UUID: UUID =
                UUID.fromString("00001827-0000-1000-8000-00805F9B34FB")

            /**
             * Mesh provisioning data in characteristic UUID
             */
            val MESH_PROVISIONING_DATA_IN: UUID =
                UUID.fromString("00002ADB-0000-1000-8000-00805F9B34FB")

            /**
             * Mesh provisioning data out characteristic UUID
             */
            val MESH_PROVISIONING_DATA_OUT: UUID =
                UUID.fromString("00002ADC-0000-1000-8000-00805F9B34FB")

            /**
             * Mesh provisioning service UUID
             */
            val MESH_PROXY_UUID: UUID = UUID.fromString("00001828-0000-1000-8000-00805F9B34FB")

            /**
             * Mesh provisioning data in characteristic UUID
             */
            val MESH_PROXY_DATA_IN: UUID = UUID.fromString("00002ADD-0000-1000-8000-00805F9B34FB")

            /**
             * Mesh provisioning data out characteristic UUID
             */
            val MESH_PROXY_DATA_OUT: UUID = UUID.fromString("00002ADE-0000-1000-8000-00805F9B34FB")
        }
    private var mMeshProvisioningDataInCharacteristic: BluetoothGattCharacteristic? = null
    private var mMeshProvisioningDataOutCharacteristic: BluetoothGattCharacteristic? = null
    private var mMeshProxyDataInCharacteristic: BluetoothGattCharacteristic? = null
    private var mMeshProxyDataOutCharacteristic: BluetoothGattCharacteristic? = null
    var mMeshManagerApi: MeshManagerApi? = null
    var isProvisioningComplete = false
    var mDeviceToConnectTo: ExtendedBluetoothDevice? = null
    private var mUnprovisionedMeshNode: UnprovisionedMeshNode? = null
    var mIsDeviceReady = false
    init {
        mMeshManagerApi =  MeshManagerApi(context)
        mMeshManagerApi!!.setMeshManagerCallbacks(this)
        mMeshManagerApi!!.setProvisioningStatusCallbacks(this)
        mMeshManagerApi!!.setMeshStatusCallbacks(this)
        mMeshManagerApi!!.loadMeshNetwork()
    }

    override fun isRequiredServiceSupported(gatt: BluetoothGatt): Boolean {
        val meshProxyService =
            gatt.getService(MESH_PROXY_UUID)
        if (meshProxyService != null) {
            isProvisioningComplete = true
            mMeshProxyDataInCharacteristic =
                meshProxyService.getCharacteristic(MESH_PROXY_DATA_IN)
            mMeshProxyDataOutCharacteristic =
                meshProxyService.getCharacteristic(MESH_PROXY_DATA_OUT)

            return mMeshProxyDataInCharacteristic != null && mMeshProxyDataOutCharacteristic != null &&
                    hasNotifyProperty(mMeshProxyDataOutCharacteristic!!) &&
                    hasWriteNoResponseProperty(mMeshProxyDataInCharacteristic!!)
        }
        val meshProvisioningService =
            gatt.getService(MESH_PROVISIONING_UUID)
        if (meshProvisioningService != null) {
            isProvisioningComplete = false
            mMeshProvisioningDataInCharacteristic =
                meshProvisioningService.getCharacteristic(MESH_PROVISIONING_DATA_IN)
            mMeshProvisioningDataOutCharacteristic =
                meshProvisioningService.getCharacteristic(MESH_PROVISIONING_DATA_OUT)

            return mMeshProvisioningDataInCharacteristic != null && mMeshProvisioningDataOutCharacteristic != null &&
                    hasNotifyProperty(mMeshProvisioningDataOutCharacteristic!!) &&
                    hasWriteNoResponseProperty(mMeshProvisioningDataInCharacteristic!!)
        }
        return false
    }

    override fun initialize() {
        requestMtu(MTU_SIZE_MAX).enqueue()
        // This callback will be called each time a notification is received.
        val onDataReceived =
            DataReceivedCallback { device: BluetoothDevice?, data: Data ->
                data.value?.let { mMeshManagerApi!!.handleNotifications(getMaximumPacketSize(), it)}
            }


        // Set the notification callback and enable notification on Data In characteristic.
        val characteristic =
            if (isProvisioningComplete) mMeshProxyDataOutCharacteristic!! else mMeshProvisioningDataOutCharacteristic!!
        setNotificationCallback(characteristic).with(onDataReceived)
        enableNotifications(characteristic).enqueue()
    }

    override fun onServicesInvalidated() {
        //We reset the MTU to 23 upon disconnection
        overrideMtu(MTU_SIZE_DEFAULT)
        mIsDeviceReady = false
        isProvisioningComplete = false
        mMeshProvisioningDataInCharacteristic = null
        mMeshProvisioningDataOutCharacteristic = null
        mMeshProxyDataInCharacteristic = null
        mMeshProxyDataOutCharacteristic = null
    }

    override fun onDeviceReady() {
        mIsDeviceReady = true
    }

    fun sendPdu(pdu: ByteArray?) {
        if (!mIsDeviceReady) return
        // This callback will be called each time the data were sent.
        val callback =
            DataSentCallback { device: BluetoothDevice?, data: Data ->
                data.value?.let { mMeshManagerApi!!.handleWriteCallbacks(getMaximumPacketSize(), it)
                    Log.d("MeshManager", "Wrote data pdu"+data.value.contentToString())
                }
            }
        // Write the right characteristic.
        val characteristic =
            if (isProvisioningComplete) mMeshProxyDataInCharacteristic!! else mMeshProvisioningDataInCharacteristic!!
        writeCharacteristic(characteristic, pdu, BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE)
            .split()
            .with(callback)
            .enqueue()

    }

    fun getMaximumPacketSize(): Int {
        return super.getMtu() - 3
    }

    private fun hasWriteNoResponseProperty(characteristic: BluetoothGattCharacteristic): Boolean {
        return (characteristic.properties and BluetoothGattCharacteristic.PROPERTY_WRITE_NO_RESPONSE) != 0
    }

    private fun hasNotifyProperty(characteristic: BluetoothGattCharacteristic): Boolean {
        return (characteristic.properties and BluetoothGattCharacteristic.PROPERTY_NOTIFY) != 0
    }


    fun identifyNode(device: ExtendedBluetoothDevice) {
        val beacon = device.beacon as UnprovisionedBeacon?
        if (beacon != null) {
            mMeshManagerApi!!.identifyNode(
                beacon.uuid,
                5
            )
        } else {
            val serviceData: ByteArray? =
                getServiceData(device.scanResult, MESH_PROVISIONING_UUID)
            if (serviceData != null) {
                val uuid = mMeshManagerApi!!.getDeviceUuid(serviceData)
                mMeshManagerApi!!.identifyNode(uuid)
            }
        }
    }

    fun getServiceData(result: ScanResult, serviceUuid: UUID): ByteArray? {
        val scanRecord = result.scanRecord
        if (scanRecord != null) {
            return scanRecord.getServiceData(ParcelUuid((serviceUuid)))
        }
        return null
    }

    val mConnectCallback: SuccessCallback = SuccessCallback { device: BluetoothDevice ->
        Log.d("MeshManager", "connected to " + device.name)
        (context as MainActivity).onDeviceConnected(device.name)
    }

    fun connectToMeshDevice(result: ScanResult) {
        Log.d("MeshManager", "Set temp device")
        mDeviceToConnectTo = ExtendedBluetoothDevice(result)
        connect(result.device).done(mConnectCallback).enqueue()
    }

    fun identifyDevice() {
        Log.d("MeshManager", "identifyDevice")
        identifyNode(mDeviceToConnectTo!!)
    }

    fun provisionDevice() {
        Log.d("MeshManager", "provisioning")

        mUnprovisionedMeshNode?.let { mMeshManagerApi!!.startProvisioning(it) }
        Log.d("MeshManager", "provisioned")
    }

    override fun getMtu() : Int {
        return  super.getMtu()
    }


    override fun onNetworkLoaded(meshNetwork: MeshNetwork?) {
        Log.d("MeshManager", "onNetworkLoaded")
    }

    override fun onNetworkUpdated(meshNetwork: MeshNetwork?) {
        Log.d("MeshManager", "onNetworkUpdated")
    }

    override fun onNetworkLoadFailed(error: String?) {
        Log.d("MeshManager", "onNetworkLoadFailed")
    }

    override fun onNetworkImported(meshNetwork: MeshNetwork?) {
        Log.d("MeshManager", "onNetworkImported")
    }

    override fun onNetworkImportFailed(error: String?) {
        Log.d("MeshManager", "onNetworkImportFailed")
    }

    override fun sendProvisioningPdu(meshNode: UnprovisionedMeshNode?, pdu: ByteArray?) {
        sendPdu(pdu)
        Log.d("MeshManager", "sendProvisioningPdu")
    }

    override fun onMeshPduCreated(pdu: ByteArray?) {
        Log.d("MeshManager", "onMeshPduCreated")
    }

    override fun onProvisioningStateChanged(
        meshNode: UnprovisionedMeshNode?,
        state: ProvisioningState.States?,
        data: ByteArray?
    ) {
        mUnprovisionedMeshNode = meshNode
        Log.d("MeshManager", "onProvisioningStateChanged")
    }

    override fun onProvisioningFailed(
        meshNode: UnprovisionedMeshNode?,
        state: ProvisioningState.States?,
        data: ByteArray?
    ) {
        Log.d("MeshManager", "onProvisioningFailed")
    }

    override fun onProvisioningCompleted(
        meshNode: ProvisionedMeshNode?,
        state: ProvisioningState.States?,
        data: ByteArray?
    ) {
        Log.d("MeshManager", "onProvisioningCompleted")
    }

    override fun onTransactionFailed(dst: Int, hasIncompleteTimerExpired: Boolean) {
        Log.d("MeshManager", "onTransactionFailed")
    }

    override fun onUnknownPduReceived(src: Int, accessPayload: ByteArray?) {
        Log.d("MeshManager", "onUnknownPduReceived")
    }

    override fun onBlockAcknowledgementProcessed(dst: Int, message: ControlMessage) {
        Log.d("MeshManager", "onBlockAcknowledgementProcessed")
    }

    override fun onBlockAcknowledgementReceived(src: Int, message: ControlMessage) {
        Log.d("MeshManager", "onBlockAcknowledgementReceived")
    }

    override fun onMeshMessageProcessed(dst: Int, meshMessage: MeshMessage) {
        Log.d("MeshManager", "onMeshMessageProcessed")
    }

    override fun onMeshMessageReceived(src: Int, meshMessage: MeshMessage) {
        Log.d("MeshManager", "onMeshMessageReceived")
    }

    override fun onMessageDecryptionFailed(meshLayer: String?, errorMessage: String?) {
        Log.d("MeshManager", "onMessageDecryptionFailed")
    }
}